from aicage._logging import get_logger
from aicage.constants import IMAGE_REGISTRY, IMAGE_REPOSITORY
from aicage.docker.pull import run_pull
from aicage.docker.query import cleanup_old_digest, get_local_repo_digest_for_repo
from aicage.registry._logs import pull_log_path
from aicage.registry._pull_decision import decide_pull
from aicage.registry._signature import resolve_verified_digest


def pull_image(image_ref: str) -> None:
    logger = get_logger()
    repository = f"{IMAGE_REGISTRY}/{IMAGE_REPOSITORY}"
    local_digest = get_local_repo_digest_for_repo(image_ref, repository)
    should_pull = decide_pull(image_ref)
    if not should_pull:
        logger.info("Image pull not required for %s", image_ref)
        return

    resolve_verified_digest(image_ref)
    log_path = pull_log_path(image_ref)
    run_pull(image_ref, log_path)
    cleanup_old_digest(repository, local_digest, image_ref)



import shlex
import subprocess
from pathlib import Path

from aicage._proxy import proxy_run_env_args_from_host
from aicage.constants import BUILDER_VERSION_CHECK_TIMEOUT_SECONDS
from aicage.docker._env import resolve_user_ids
from aicage.docker._mounts import append_mount
from aicage.docker.cli import run_docker_command
from aicage.docker.query import cleanup_old_digest, get_local_repo_digest_for_repo
from aicage.docker.refs import repository_from_image_ref
from aicage.runtime.run_args import DockerRunArgs


def run_container(args: DockerRunArgs) -> None:
    command = _assemble_docker_run(args)
    repository = repository_from_image_ref(args.image_ref)
    old_digest = get_local_repo_digest_for_repo(args.image_ref, repository)
    try:
        run_docker_command(command, check=True)
    finally:
        cleanup_old_digest(repository, old_digest, args.image_ref)


def print_run_command(args: DockerRunArgs) -> None:
    command = _assemble_docker_run(args)
    print(shlex.join(command))


def run_builder_version_check(image_ref: str, definition_dir: Path) -> subprocess.CompletedProcess[str]:
    command = [
        "/bin/bash",
        "-c",
        "cp /agent/version.sh /tmp/version.sh "
        "&& sed -i 's/\\r$//' /tmp/version.sh "
        "&& chmod +x /tmp/version.sh "
        "&& /bin/bash /tmp/version.sh",
    ]
    run_command = [
        "docker",
        "run",
        "--rm",
        *proxy_run_env_args_from_host(),
        "-v",
        f"{str(definition_dir.resolve())}:/agent:ro",
        "-w",
        "/agent",
        image_ref,
        *command,
    ]
    try:
        process = subprocess.run(
            run_command,
            check=False,
            capture_output=True,
            text=True,
            timeout=BUILDER_VERSION_CHECK_TIMEOUT_SECONDS,
        )
        return subprocess.CompletedProcess(
            command,
            process.returncode,
            stdout=process.stdout,
            stderr=process.stderr,
        )
    except subprocess.TimeoutExpired:
        return subprocess.CompletedProcess(command, 124, stdout="", stderr="Version check timed out.")
    except Exception as exc:
        return subprocess.CompletedProcess(command, 1, stdout="", stderr=str(exc))



def _assemble_docker_run(args: DockerRunArgs) -> list[str]:
    cmd: list[str] = ["docker", "run", "--rm", "-it"]
    cmd.extend(resolve_user_ids())
    for env in args.env:
        cmd.extend(["-e", f"{env.name}={env.value}"])
    for mount in args.mounts:
        append_mount(cmd, mount.host_path, mount.container_path, read_only=mount.read_only)

    if args.merged_docker_args:
        cmd.extend(shlex.split(args.merged_docker_args))

    cmd.append(args.image_ref)
    cmd.extend(args.agent_args)
    return cmd



import os
from pathlib import Path

from aicage._logging import get_logger
from aicage.runtime._errors import RuntimeExecutionError

from ._tty import ensure_tty_for_prompt
from .mode import assume_yes_enabled


def _prompt_yes_no(question: str, default: bool = False) -> bool:
    if assume_yes_enabled():
        get_logger().info("Prompt yes/no '%s' -> %s (assume-yes)", question, default)
        return default
    ensure_tty_for_prompt()
    suffix = "[Y/n]" if default else "[y/N]"
    response = input(f"{question} {suffix} ").strip().lower()
    if not response:
        choice = default
    else:
        choice = response in {"y", "yes"}
    get_logger().info("Prompt yes/no '%s' -> %s", question, choice)
    return choice


def prompt_persist_docker_socket() -> bool:
    if os.name == "nt":
        print(
            "Info: You must enable 'Expose daemon on tcp://localhost:2375 without TLS' "
            "in Docker Desktop settings to use --docker on Windows."
        )
    return _prompt_yes_no("Persist mounting the Docker socket for this project?", default=True)


def prompt_mount_git_support(items: list[tuple[str, str, Path]]) -> list[str]:
    if assume_yes_enabled():
        selected_keys = [item[0] for item in items]
        get_logger().info("Prompt git support mounts selected -> %s (assume-yes)", selected_keys)
        return selected_keys
    ensure_tty_for_prompt()
    logger = get_logger()
    print("Enable Git support in the container by mounting:")
    for idx, (_, label, path) in enumerate(items, start=1):
        print(f"  {idx}) {label}: {path}")
    response = input("Select mounts (comma-separated numbers) [all, default on Enter]: ").strip()
    if not response:
        selected = set(range(1, len(items) + 1))
    else:
        selected = _parse_number_selection(response, len(items))
    selected_keys = [items[idx - 1][0] for idx in sorted(selected)]
    logger.info("Prompt git support mounts selected -> %s", selected_keys)
    return selected_keys


def _parse_number_selection(response: str, max_value: int) -> set[int]:
    selected: set[int] = set()
    tokens = [item.strip() for item in response.split(",") if item.strip()]
    if not tokens:
        raise RuntimeExecutionError("Invalid selection ''. Provide comma-separated numbers, e.g. 1,2.")
    for token in tokens:
        if not token.isdigit():
            raise RuntimeExecutionError(f"Invalid selection '{token}'. Use numbers between 1 and {max_value}.")
        value = int(token)
        if value < 1 or value > max_value:
            raise RuntimeExecutionError(
                f"Invalid selection '{token}'. Pick a number between 1 and {max_value}."
            )
        if value in selected:
            raise RuntimeExecutionError(f"Duplicate selection '{token}' is not allowed.")
        selected.add(value)
    return selected


def prompt_persist_docker_args(new_args: str, existing_args: str | None) -> bool:
    if existing_args:
        question = f"Persist docker run args '{new_args}' for this project (replacing '{existing_args}')?"
    else:
        question = f"Persist docker run args '{new_args}' for this project?"
    return _prompt_yes_no(question, default=True)


def prompt_persist_shares(new_shares: list[str], existing_shares: list[str]) -> bool:
    new_values = ", ".join(new_shares)
    if existing_shares:
        question = (
            "Persist share mounts "
            f"'{new_values}' for this project (adding to {len(existing_shares)} existing share(s))?"
        )
    else:
        question = f"Persist share mounts '{new_values}' for this project?"
    return _prompt_yes_no(question, default=True)


def prompt_update_aicage(installed_version: str, latest_version: str) -> bool:
    question = (
        "A newer version of aicage is available "
        f"(installed: {installed_version}, latest: {latest_version}). "
        "Update now?"
    )
    return _prompt_yes_no(question, default=True)



from pathlib import Path

from aicage.config.agent._custom_loader import load_custom_agents
from aicage.config.agent._loader_shared import load_agents_from_directory
from aicage.config.agent.models import AgentMetadata
from aicage.config.base.models import BaseMetadata
from aicage.config.errors import ConfigError
from aicage.config.resources import find_packaged_path

_AGENT_DEFINITION_FILES: tuple[str, str] = ("agent.yaml", "agent.yml")


def load_agents(bases: dict[str, BaseMetadata]) -> dict[str, AgentMetadata]:
    builtin_agents = _load_builtin_agents(bases)
    custom_agents = load_custom_agents(bases)
    merged_agents = dict(builtin_agents)
    merged_agents.update(custom_agents)
    return merged_agents


def _load_builtin_agents(bases: dict[str, BaseMetadata]) -> dict[str, AgentMetadata]:
    agents_dir = _builtin_agents_dir()
    if not agents_dir.is_dir():
        raise ConfigError(f"Built-in agent directory '{agents_dir}' is missing.")
    return load_agents_from_directory(
        agents_dir=agents_dir,
        bases=bases,
        definition_files=_AGENT_DEFINITION_FILES,
        agent_label="Agent",
    )


def _builtin_agents_dir() -> Path:
    dockerfile = find_packaged_path("agent-build/Dockerfile")
    return dockerfile.parent / "agents"



from itertools import chain
from pathlib import Path

from aicage.cli_types import ParsedArgs
from aicage.config.context import ConfigContext
from aicage.config.project_config import AgentConfig
from aicage.errors import AicageError
from aicage.paths import container_project_path
from aicage.runtime.docker_args._support._resolver_types import MountRequest, ResolvedArgs, Resolver
from aicage.runtime.env_vars import AICAGE_WORKSPACE
from aicage.runtime.run_args import EnvVar, MountSpec

from .._resolvers import _agent_config, _docker_socket, _git_config, _git_root, _gpg, _project, _shares, _ssh_keys
from .._support._git_support import resolve_git_support_prefs
from ._mounts import map_mount_requests


def resolve_docker_args(
    context: ConfigContext,
    agent: str,
    parsed: ParsedArgs | None,
) -> tuple[list[MountSpec], list[EnvVar]]:
    agent_cfg = context.project_cfg.agents.setdefault(agent, AgentConfig())
    project_path = Path(context.project_cfg.path).resolve()
    resolve_git_support_prefs(project_path, agent_cfg)
    resolved = [_resolve_provider(provider, context, agent, parsed) for provider in _resolver_sequence()]
    mount_requests = list(chain.from_iterable(item.mounts for item in resolved))
    env = list(chain.from_iterable(item.env for item in resolved))
    host_home = Path.home().resolve()
    mounts = _map_mount_requests(mount_requests)
    _validate_home_mount_safety(mounts, host_home)
    workspace_path = container_project_path(project_path)
    env.append(EnvVar(name=AICAGE_WORKSPACE, value=workspace_path.as_posix()))
    return mounts, env


def _resolve_provider(
    provider: Resolver,
    context: ConfigContext,
    agent: str,
    parsed: ParsedArgs | None,
) -> ResolvedArgs:
    return provider(context, agent, parsed)


def _resolver_sequence() -> tuple[Resolver, ...]:
    return (
        _project.resolve,
        _agent_config.resolve,
        _git_config.resolve,
        _gpg.resolve,
        _ssh_keys.resolve,
        _git_root.resolve,
        _docker_socket.resolve,
        _shares.resolve,
    )


def _map_mount_requests(
    requests: list[MountRequest],
) -> list[MountSpec]:
    return map_mount_requests(requests)


def _validate_home_mount_safety(mounts: list[MountSpec], host_home: Path) -> None:
    for mount in mounts:
        host_path = mount.host_path.resolve()
        if host_path == host_home or host_path in host_home.parents:
            raise AicageError(
                "Refusing to start: this would mount your home directory into the container: "
                f"{host_path}",
            )



from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from typing import Protocol

from aicage.config.project_config import AgentConfig
from aicage.paths import HOST_GNUPG_DIR, HOST_SSH_DIR

from ...prompts.confirm import prompt_mount_git_support
from ._exec import capture_stdout
from ._signing import is_commit_signing_enabled, resolve_signing_format

_REMOTE_WITH_URL_FIELD_COUNT: int = 2
_REMOTE_URL_INDEX: int = 1


@dataclass(frozen=True)
class _GitSupportPromptItem:
    key: str
    label: str
    path: Path


class _MountPrefs(Protocol):
    gitconfig: bool | None
    gitroot: bool | None
    gnupg: bool | None
    ssh: bool | None


def resolve_git_config_path() -> Path | None:
    stdout = capture_stdout(["git", "config", "--global", "--show-origin", "--list"])
    if not stdout:
        return None
    for line in stdout.splitlines():
        if not line.startswith("file:"):
            continue
        parts = line[5:].split()
        if not parts:
            continue
        return Path(parts[0]).expanduser()
    return None


def resolve_git_root(project_path: Path) -> Path | None:
    superproject = _resolve_git_path(
        ["git", "rev-parse", "--show-superproject-working-tree"],
        project_path,
    )
    if superproject:
        return superproject
    return _resolve_git_path(["git", "rev-parse", "--show-toplevel"], project_path)


def _resolve_git_path(command: list[str], project_path: Path) -> Path | None:
    stdout = capture_stdout(command, cwd=project_path)
    if not stdout:
        return None
    value = stdout.strip()
    if not value:
        return None
    return Path(value).resolve()


def resolve_gpg_home() -> Path | None:
    stdout = capture_stdout(["gpgconf", "--list-dirs", "homedir"])
    if stdout:
        path = stdout.strip()
        if path:
            gpg_home = Path(path).expanduser()
            if gpg_home.exists():
                return gpg_home
    fallback = HOST_GNUPG_DIR
    return fallback if fallback.exists() else None


def resolve_ssh_dir() -> Path:
    return HOST_SSH_DIR


def uses_ssh_remotes(project_path: Path) -> bool:
    stdout = capture_stdout(["git", "remote", "-v"], cwd=project_path)
    if not stdout:
        return False
    for line in stdout.splitlines():
        fields = line.split()
        if len(fields) < _REMOTE_WITH_URL_FIELD_COUNT:
            continue
        remote_url = fields[_REMOTE_URL_INDEX]
        if remote_url.startswith("ssh://") or "@" in remote_url:
            return True
    return False


def resolve_git_support_prefs(project_path: Path, agent_cfg: AgentConfig) -> None:
    mounts_cfg = agent_cfg.mounts
    items: list[_GitSupportPromptItem] = []

    git_config = resolve_git_config_path()
    if git_config and git_config.exists() and mounts_cfg.gitconfig is None:
        items.append(_GitSupportPromptItem("gitconfig", "Git config (name/email)", git_config))

    git_root = resolve_git_root(project_path)
    if git_root and git_root != project_path and mounts_cfg.gitroot is None:
        items.append(_GitSupportPromptItem("gitroot", "Git root (repository access)", git_root))

    signing_enabled = is_commit_signing_enabled(project_path)
    signing_format = resolve_signing_format(project_path) if signing_enabled else None
    ssh_needed = (signing_enabled and signing_format == "ssh") or uses_ssh_remotes(project_path)

    if ssh_needed and mounts_cfg.ssh is None:
        ssh_dir = resolve_ssh_dir()
        if ssh_dir.exists():
            items.append(
                _GitSupportPromptItem(
                    "ssh",
                    "SSH keys (for Git SSH/signing)",
                    ssh_dir,
                )
            )

    needs_gnupg = signing_enabled and signing_format != "ssh"
    if needs_gnupg and mounts_cfg.gnupg is None:
        gpg_home = resolve_gpg_home()
        if gpg_home and gpg_home.exists():
            items.append(
                _GitSupportPromptItem(
                    "gnupg",
                    "GnuPG keys (for Git signing)",
                    gpg_home,
                )
            )

    if not items:
        return

    selected = set(prompt_mount_git_support(_format_prompt_items(items)))
    for item in items:
        _set_mount_pref(mounts_cfg, item.key, item.key in selected)


def _format_prompt_items(items: Iterable[_GitSupportPromptItem]) -> list[tuple[str, str, Path]]:
    return [(item.key, item.label, item.path) for item in items]


def _set_mount_pref(mounts_cfg: _MountPrefs, key: str, value: bool) -> None:
    setattr(mounts_cfg, key, value)



import os

from aicage.cli_types import ParsedArgs
from aicage.config.context import ConfigContext
from aicage.config.project_config import AgentConfig
from aicage.paths import HOST_DOCKER_SOCKET_PATH
from aicage.runtime.docker_args._support._resolver_types import MountRequest, ResolvedArgs
from aicage.runtime.env_vars import DOCKER_HOST, WINDOWS_DOCKER_HOST
from aicage.runtime.prompts.confirm import prompt_persist_docker_socket
from aicage.runtime.run_args import EnvVar


def resolve(
    context: ConfigContext,
    agent: str,
    parsed: ParsedArgs | None,
) -> ResolvedArgs:
    agent_cfg: AgentConfig = context.project_cfg.agents[agent]
    mounts_cfg = agent_cfg.mounts
    cli_docker_socket = parsed.docker_socket if parsed else False
    docker_socket_enabled = cli_docker_socket or bool(mounts_cfg.docker)
    if not docker_socket_enabled:
        return ResolvedArgs()

    if os.name == "nt":
        mounts: list[MountRequest] = []
        env = [EnvVar(name=DOCKER_HOST, value=WINDOWS_DOCKER_HOST)]
    else:
        mounts = [MountRequest(host_path=HOST_DOCKER_SOCKET_PATH)]
        env = []

    if cli_docker_socket and mounts_cfg.docker is None:
        if prompt_persist_docker_socket():
            mounts_cfg.docker = True

    return ResolvedArgs(mounts=mounts, env=env)



def is_base_excluded(
    base_name: str,
    base_distro: str,
    base_exclude: set[str],
    base_distro_exclude: set[str],
) -> bool:
    base_name_lc = base_name.lower()
    if base_name_lc in base_exclude:
        return True
    if base_distro.lower() in base_distro_exclude:
        return True
    return False


def normalize_exclude(values: list[str]) -> set[str]:
    if not values:
        return set()
    return {value.lower() for value in values}



from pathlib import Path

from aicage.cli_types import ParsedArgs
from aicage.config.context import ConfigContext
from aicage.config.project_config import AgentConfig
from aicage.runtime.docker_args._support._resolver_types import MountRequest, ResolvedArgs

from .._support._git_support import resolve_git_root


def resolve(
    context: ConfigContext,
    agent: str,
    parsed: ParsedArgs | None,
) -> ResolvedArgs:
    _ = parsed
    project_path = Path(context.project_cfg.path)
    agent_cfg: AgentConfig = context.project_cfg.agents[agent]
    git_root = resolve_git_root(project_path)
    if not git_root or git_root == project_path:
        return ResolvedArgs()

    mounts_cfg = agent_cfg.mounts
    if mounts_cfg.gitroot:
        return ResolvedArgs(mounts=[MountRequest(host_path=git_root)])
    return ResolvedArgs()



_PROMPT_MODE: dict[str, bool] = {"assume_yes": False}


def set_assume_yes(value: bool) -> None:
    _PROMPT_MODE["assume_yes"] = value


def assume_yes_enabled() -> bool:
    return _PROMPT_MODE["assume_yes"]



from pathlib import Path

from aicage.cli_types import ParsedArgs
from aicage.config.context import ConfigContext
from aicage.runtime.docker_args._support._resolver_types import MountRequest, ResolvedArgs
from aicage.runtime.mounts.shares import ShareSpec, resolve_share_mounts


def resolve(
    context: ConfigContext,
    agent: str,
    parsed: ParsedArgs | None,
) -> ResolvedArgs:
    _ = (context, agent)
    if parsed is None:
        return ResolvedArgs()
    cwd = Path.cwd().resolve()
    share_mounts: list[ShareSpec] = resolve_share_mounts(parsed, cwd)
    mounts = [MountRequest(host_path=share.host_path, read_only=share.read_only) for share in share_mounts]
    return ResolvedArgs(mounts=mounts)



from aicage.cli_types import ParsedArgs
from aicage.config.context import ConfigContext
from aicage.runtime._agent_config import AgentConfig, resolve_agent_config
from aicage.runtime.docker_args._support._resolver_types import MountRequest, ResolvedArgs


def resolve(
    context: ConfigContext,
    agent: str,
    parsed: ParsedArgs | None,
) -> ResolvedArgs:
    _ = parsed
    agent_metadata = context.agents[agent]
    agent_config: AgentConfig = resolve_agent_config(agent_metadata)
    mounts = [MountRequest(host_path=host_path) for host_path in agent_config.agent_config_host]
    return ResolvedArgs(mounts=mounts)



from pathlib import Path

from aicage.cli_types import ParsedArgs
from aicage.config.context import ConfigContext
from aicage.config.project_config import AgentConfig
from aicage.runtime.docker_args._support._resolver_types import MountRequest, ResolvedArgs

from .._support._git_support import resolve_gpg_home
from .._support._signing import is_commit_signing_enabled, resolve_signing_format


def resolve(
    context: ConfigContext,
    agent: str,
    parsed: ParsedArgs | None,
) -> ResolvedArgs:
    _ = parsed
    project_path = Path(context.project_cfg.path)
    agent_cfg: AgentConfig = context.project_cfg.agents[agent]
    if not is_commit_signing_enabled(project_path):
        return ResolvedArgs()
    if resolve_signing_format(project_path) == "ssh":
        return ResolvedArgs()

    gpg_home = resolve_gpg_home()
    if not gpg_home or not gpg_home.exists():
        return ResolvedArgs()

    mounts_cfg = agent_cfg.mounts
    if mounts_cfg.gnupg:
        return ResolvedArgs(mounts=[MountRequest(host_path=gpg_home)])
    return ResolvedArgs()



IMAGE_REGISTRY: str = "ghcr.io"
_IMAGE_REGISTRY_API_URL: str = "https://ghcr.io/v2"
_IMAGE_REGISTRY_API_TOKEN_URL: str = (
    "https://ghcr.io/token?service=ghcr.io&scope=repository"
)
IMAGE_REPOSITORY: str = "aicage/aicage"
IMAGE_BASE_REPOSITORY: str = "aicage/aicage-image-base"
DEFAULT_IMAGE_BASE: str = "ubuntu"
VERSION_CHECK_IMAGE: str = "ghcr.io/aicage/aicage-image-util:agent-version"

LOCAL_IMAGE_BASE_REPOSITORY: str = "aicage-image-base"
LOCAL_IMAGE_REPOSITORY: str = "aicage"

DEFAULT_EXTENDED_IMAGE_NAME: str = "aicage-extended"

_COSIGN_IMAGE_NAME: str = "ghcr.io/sigstore/cosign/cosign"
# _COSIGN_IMAGE_DIGEST holds the digest of
# ghcr.io/sigstore/cosign/cosign:latest
# at the time of a release
_COSIGN_IMAGE_DIGEST: str = "sha256:be924970ba7438c22e18067dec5637946d6566eac711f5bedd1584e7137008fb"
COSIGN_IMAGE_REF: str = f"{_COSIGN_IMAGE_NAME}@{_COSIGN_IMAGE_DIGEST}"
COSIGN_OIDC_ISSUER: str = "https://token.actions.githubusercontent.com"
COSIGN_IDENTITY_REGEXP: str = (
    "^https://github.com/aicage/[^/]+/.github/workflows/.*@(?:refs/.*/.*|[0-9a-f]{40})$"
)

DOCKER_LOCAL_METADATA_TIMEOUT_SECONDS: int = 30
DOCKER_PULL_REQUEST_TIMEOUT_SECONDS: int = 60 * 60 * 6
DOCKER_REGISTRY_REQUEST_TIMEOUT_SECONDS: float = 5.0
REGISTRY_DIGEST_REQUEST_TIMEOUT_SECONDS: float = 4.0
PYPI_VERSION_CHECK_TIMEOUT_SECONDS: float = 3.5
PROJECT_FILE_LOCK_TIMEOUT_SECONDS: int = 30
HOST_VERSION_CHECK_TIMEOUT_SECONDS: float = 15.0
BUILDER_VERSION_CHECK_TIMEOUT_SECONDS: float = 20.0
PROXY_ENV_NAMES: tuple[str, ...] = (
    "HTTP_PROXY",
    "HTTPS_PROXY",
    "ALL_PROXY",
    "NO_PROXY",
)



import os
import subprocess
import sys
from collections.abc import Sequence

from aicage import __version__
from aicage._logging import get_logger
from aicage.cli._info_config import info_project_config
from aicage.cli._parse import parse_cli
from aicage.cli._remove_config import remove_project_config
from aicage.cli._version_check import maybe_prompt_update
from aicage.cli_types import ParsedArgs
from aicage.config.runtime_config import RunConfig, load_run_config
from aicage.docker.errors import DockerError
from aicage.docker.run import print_run_command, run_container
from aicage.errors import AicageError
from aicage.registry.ensure_image import ensure_image
from aicage.runtime.prompts.mode import set_assume_yes
from aicage.runtime.run_args import DockerRunArgs
from aicage.runtime.run_plan import build_run_args


def main(argv: Sequence[str] | None = None) -> int:
    parsed_argv: Sequence[str] = argv if argv is not None else sys.argv[1:]
    logger = get_logger()
    exit_code = 0
    try:
        parsed: ParsedArgs = parse_cli(parsed_argv)
        set_assume_yes(parsed.yes)
        if maybe_prompt_update(__version__):
            _restart_with_current_args(parsed_argv)
        if parsed.config_action == "info":
            info_project_config()
        elif parsed.config_action == "remove":
            remove_project_config(parsed.config_agent)
        else:
            run_config: RunConfig = load_run_config(parsed.agent, parsed)
            run_args: DockerRunArgs = build_run_args(config=run_config, parsed=parsed)
            logger.info("Resolved run config for agent %s", run_config.agent)
            ensure_image(run_config)

            if parsed.dry_run:
                print_run_command(run_args)
                logger.info("Dry-run docker command printed.")
            else:
                run_container(run_args)
    except KeyboardInterrupt:
        print()
        logger.warning("Interrupted by user.")
        exit_code = 130
    except DockerError as exc:
        _print_docker_error(exc)
        logger.error("CLI error: %s", exc)
        exit_code = 1
    except AicageError as exc:
        print(f"[aicage] {exc}", file=sys.stderr)
        logger.error("CLI error: %s", exc)
        exit_code = 1
    return exit_code


def _print_docker_error(exc: DockerError) -> None:
    print(f"[aicage] {exc}", file=sys.stderr)
    cause = exc.__cause__
    if not isinstance(cause, subprocess.CalledProcessError):
        return
    stderr = cause.stderr
    if isinstance(stderr, bytes):
        stderr = stderr.decode("utf-8", errors="replace")
    if stderr:
        print(stderr.rstrip(), file=sys.stderr)


def _restart_with_current_args(parsed_argv: Sequence[str]) -> None:
    os.execv(sys.executable, [sys.executable, "-m", "aicage", *parsed_argv])



from aicage.config.runtime_config import RunConfig
from aicage.docker.query import local_image_exists
from aicage.registry._layers import base_layer_missing

from ._store import BuildRecord


def should_rebuild(
    run_config: RunConfig,
    record: BuildRecord | None,
    agent_version: str,
    base_image_ref: str,
) -> bool:
    image_ref = run_config.selection.base_image_ref
    if not local_image_exists(image_ref):
        return True
    if record is None:
        return True
    if record.agent_version != agent_version:
        return True
    return base_layer_missing(base_image_ref, image_ref)



import subprocess

from aicage._logging import get_logger
from aicage.constants import COSIGN_IDENTITY_REGEXP, COSIGN_IMAGE_REF, COSIGN_OIDC_ISSUER
from aicage.docker.cli import run_docker_command_capture
from aicage.docker.pull import run_pull
from aicage.docker.query import (
    cleanup_old_digest,
    get_local_repo_digest_for_repo,
    local_image_exists,
)
from aicage.registry._errors import RegistryError
from aicage.registry._logs import pull_log_path
from aicage.registry.digest.remote_digest import get_remote_digest


def resolve_verified_digest(image_ref: str) -> str:
    logger = get_logger()
    digest = get_remote_digest(image_ref)
    if digest is None:
        raise RegistryError(f"Failed to resolve remote digest for {image_ref}.")

    digest_ref = _with_digest(image_ref, digest)
    _ensure_cosign_image()
    logger.info("Verifying image signature for %s", digest_ref)
    result = _run_cosign_verify(digest_ref)
    output = _format_cosign_output(result)
    if result.returncode == 0:
        if output:
            logger.info("Image signature verification output for %s:\n%s", digest_ref, output)
        logger.info("Image signature verification succeeded for %s", digest_ref)
        return digest_ref
    if output:
        logger.error("Image signature verification output for %s:\n%s", digest_ref, output)
    raise RegistryError(
        "Image signature verification failed for "
        f"{digest_ref}.\nCosign output:\n{output}"
    )


def _with_digest(image_ref: str, digest: str) -> str:
    name = image_ref.split("@", 1)[0]
    if ":" in name and name.rfind(":") > name.rfind("/"):
        name = name[: name.rfind(":")]
    return f"{name}@{digest}"


def _run_cosign_verify(image_ref: str) -> subprocess.CompletedProcess[str]:
    command = [
        "docker",
        "run",
        "--rm",
        COSIGN_IMAGE_REF,
        "verify",
        "--certificate-oidc-issuer",
        COSIGN_OIDC_ISSUER,
        "--certificate-identity-regexp",
        COSIGN_IDENTITY_REGEXP,
        image_ref,
    ]
    return run_docker_command_capture(
        command,
        check=False,
        text=True,
    )


def _format_cosign_output(result: subprocess.CompletedProcess[str]) -> str:
    parts = [result.stdout.strip(), result.stderr.strip()]
    return "\n".join(part for part in parts if part)


def _ensure_cosign_image() -> None:
    logger = get_logger()
    repository = _repository_for_image(COSIGN_IMAGE_REF)
    local_digest = get_local_repo_digest_for_repo(COSIGN_IMAGE_REF, repository)
    if local_image_exists(COSIGN_IMAGE_REF):
        logger.info("Cosign image already present: %s", COSIGN_IMAGE_REF)
        return
    logger.info("Pulling cosign image %s for signature verification", COSIGN_IMAGE_REF)
    log_path = pull_log_path(COSIGN_IMAGE_REF)
    run_pull(COSIGN_IMAGE_REF, log_path)
    cleanup_old_digest(repository, local_digest, COSIGN_IMAGE_REF)


def _repository_for_image(image_ref: str) -> str:
    name = image_ref.split("@", 1)[0]
    last_colon = name.rfind(":")
    if last_colon > name.rfind("/"):
        return name[:last_colon]
    return name



from pathlib import Path, PurePosixPath


def append_mount(
    cmd: list[str],
    host_path: Path,
    container_path: PurePosixPath,
    *,
    read_only: bool,
) -> None:
    mount_value = _format_mount_value(host_path, container_path, read_only=read_only)
    cmd.extend(["--mount", mount_value])


def _format_mount_value(
    host_path: Path,
    container_path: PurePosixPath,
    *,
    read_only: bool,
) -> str:
    parts = [
        "type=bind",
        f"src={host_path.as_posix()}",
        f"dst={container_path.as_posix()}",
    ]
    if read_only:
        parts.append("readonly")
    return ",".join(parts)



from aicage.config._base_exclude import is_base_excluded, normalize_exclude
from aicage.config.agent.models import AgentMetadata
from aicage.config.context import ConfigContext


def filter_bases(context: ConfigContext, agent_metadata: AgentMetadata) -> set[str]:
    base_exclude = normalize_exclude(agent_metadata.base_exclude)
    base_distro_exclude = normalize_exclude(agent_metadata.base_distro_exclude)
    filtered: set[str] = set()
    for base_name, base_metadata in context.bases.items():
        if is_base_excluded(
            base_name,
            base_metadata.base_image_distro,
            base_exclude,
            base_distro_exclude,
        ):
            continue
        filtered.add(base_name)
    return filtered



import os

from aicage.constants import PROXY_ENV_NAMES
from aicage.runtime.run_args import EnvVar


def proxy_env_vars_from_host() -> list[EnvVar]:
    env_vars: list[EnvVar] = []
    for name in PROXY_ENV_NAMES:
        value = os.environ.get(name)
        if value:
            env_vars.append(EnvVar(name=name, value=value))
    return env_vars


def proxy_build_args_from_host() -> list[str]:
    args: list[str] = []
    for env_var in proxy_env_vars_from_host():
        args.extend(["--build-arg", f"{env_var.name}={env_var.value}"])
    return args


def proxy_run_env_args_from_host() -> list[str]:
    args: list[str] = []
    for env_var in proxy_env_vars_from_host():
        args.extend(["-e", f"{env_var.name}={env_var.value}"])
    return args



import os
from dataclasses import dataclass
from pathlib import Path

from aicage.config.agent.models import AgentMetadata
from aicage.runtime._path_utils import ensure_path_exists


@dataclass
class AgentConfig:
    agent_path_files: list[str]
    agent_path_directories: list[str]
    agent_config_host: list[Path]


def resolve_agent_config(agent_metadata: AgentMetadata) -> AgentConfig:
    agent_paths = [
        *_ensure_agent_paths(agent_metadata.agent_path_files, True),
        *_ensure_agent_paths(agent_metadata.agent_path_directories, False),
    ]
    return AgentConfig(
        agent_path_files=list(agent_metadata.agent_path_files),
        agent_path_directories=list(agent_metadata.agent_path_directories),
        agent_config_host=agent_paths,
    )


def _ensure_agent_paths(agent_paths: list[str], is_file: bool) -> list[Path]:
    ensured: list[Path] = []
    for agent_path in agent_paths:
        expanded = Path(os.path.expanduser(agent_path)).resolve()
        ensured.append(ensure_path_exists(expanded, is_file))
    return ensured



from pathlib import Path

from aicage.paths import container_project_path
from aicage.runtime.docker_args._support._resolver_types import MountRequest
from aicage.runtime.run_args import MountSpec


def map_mount_requests(requests: list[MountRequest]) -> list[MountSpec]:
    deduped_requests = _dedupe_exact_mount_requests(requests)
    selected_requests = _dedupe_nested_mount_requests(deduped_requests)
    return _build_mount_specs(selected_requests)


def _dedupe_exact_mount_requests(requests: list[MountRequest]) -> list[MountRequest]:
    deduped_requests: dict[Path, MountRequest] = {}
    for request in requests:
        host_path = request.host_path.resolve()
        if host_path in deduped_requests:
            continue
        deduped_requests[host_path] = MountRequest(host_path=host_path, read_only=request.read_only)
    return list(deduped_requests.values())


def _dedupe_nested_mount_requests(requests: list[MountRequest]) -> list[MountRequest]:
    selected_requests: list[MountRequest] = []
    for request in requests:
        host_path = request.host_path.resolve()
        if _is_nested_path(host_path, [item.host_path for item in selected_requests]):
            continue
        selected_requests = [item for item in selected_requests if not _is_path_within(item.host_path, host_path)]
        selected_requests.append(request)
    return selected_requests


def _build_mount_specs(requests: list[MountRequest]) -> list[MountSpec]:
    mounts: list[MountSpec] = []
    for request in requests:
        host_path = request.host_path.resolve()
        container_path = container_project_path(host_path)
        mounts.append(
            MountSpec(
                host_path=host_path,
                container_path=container_path,
                read_only=request.read_only,
            )
        )
    return mounts


def _is_nested_path(path: Path, selected_hosts: list[Path]) -> bool:
    for selected in selected_hosts:
        if _is_path_within(path, selected):
            return True
    return False


def _is_path_within(path: Path, parent: Path) -> bool:
    try:
        path.relative_to(parent)
        return True
    except ValueError:
        return False



import subprocess
from typing import Literal, TextIO, overload

from aicage.docker.errors import DockerError


def run_docker_command(
    command: list[str],
    *,
    check: bool,
    stdout: TextIO | int | None = None,
    stderr: TextIO | int | None = None,
) -> subprocess.CompletedProcess[str] | subprocess.CompletedProcess[bytes]:
    try:
        return subprocess.run(
            command,
            check=check,
            stdout=stdout,
            stderr=stderr,
        )
    except FileNotFoundError as exc:
        raise DockerError("Docker CLI not found. Install Docker and ensure it is on PATH.") from exc
    except subprocess.CalledProcessError as exc:
        raise DockerError(f"Docker command failed with exit code {exc.returncode}.") from exc


@overload
def run_docker_command_capture(
    command: list[str],
    *,
    check: bool,
    text: Literal[True],
) -> subprocess.CompletedProcess[str]:
    ...


@overload
def run_docker_command_capture(
    command: list[str],
    *,
    check: bool,
    text: Literal[False],
) -> subprocess.CompletedProcess[bytes]:
    ...


def run_docker_command_capture(
    command: list[str],
    *,
    check: bool,
    text: bool,
) -> subprocess.CompletedProcess[str] | subprocess.CompletedProcess[bytes]:
    try:
        return subprocess.run(
            command,
            check=check,
            capture_output=True,
            text=text,
        )
    except FileNotFoundError as exc:
        raise DockerError("Docker CLI not found. Install Docker and ensure it is on PATH.") from exc
    except subprocess.CalledProcessError as exc:
        raise DockerError(f"Docker command failed with exit code {exc.returncode}.") from exc
